/*
  ICFS: Interactively Controlled File System
  Copyright (C) 2024-2025  Fedir Kovalov

  This program can be distributed under the terms of the GNU GPLv2.
  See the file LICENSE.
*/

#include "temp_permissions_table.h"
#include "cc.h"
#include "process_info.h"
#include <pthread.h>
#include <stdio.h>
#include <sys/types.h>

struct temp_process_permissions {
  // yes, this is a correct type for start time in jiffies (see
  // proc_pid_stat(5))
  unsigned long long creation_time;
  vec(char *) allowed_files;
};

map(pid_t, struct temp_process_permissions) temp_permissions_table;
pthread_mutex_t temp_permissions_table_lock;

/**
 * Function to get the process creation time (in jiffies) from the proc
 * filesystem
 *
 * @param pid: The process ID of the process to get the creation time of
 * @return: The process creation time in jiffies, or 0 on error
 * @note: although nothing in the documentation says that the creation time is
 * never really equal to 0, it exceptionally unlikely.
 */
unsigned long long get_process_creation_time(pid_t pid) {
  char path[32];
  FILE *fp;
  unsigned long long creation_time = 0;

  // Construct the path to the process's status file
  snprintf(path, sizeof(path), "/proc/%u/stat", pid);

  // Open the status file
  fp = fopen(path, "r");
  if (fp == NULL) {
    perror("fopen");
    return 0;
  }

  // Read the creation time (the 22nd field in the stat file)
  for (int i = 1; i < 22; i++) {
    if (fscanf(fp, "%*s") == EOF) {
      fprintf(stderr, "Error reading process stat file on the number %d\n", i);
      fclose(fp);
      return 0;
    }
  }
  if (fscanf(fp, "%llu", &creation_time) != 1) {
    fprintf(stderr, "Error reading creation time\n");
    fclose(fp);
    return 0;
  }

  // Close the file
  fclose(fp);

  return creation_time;
}

/**
 * Initializes the temporary permissions table.
 *
 * @return: 0 on success, -1 on failure (e.g. ENOMEM)
 */
int init_temp_permissions_table() {
  pthread_mutex_init(&temp_permissions_table_lock, PTHREAD_MUTEX_DEFAULT);
  init(&temp_permissions_table);
  return 0;
}

/**
 * Destroys the temporary permissions table.
 *
 * @note: the table is guranteed to be destroyed if it is already initialized
 */
void destroy_temp_permissions_table() {
  // free the memory allocated for the table
  for_each(&temp_permissions_table, entry) {
    for_each(&entry->allowed_files, allowed_file) { free(*allowed_file); }
    cleanup(&entry->allowed_files);
  }
  cleanup(&temp_permissions_table);
  pthread_mutex_destroy(&temp_permissions_table_lock);
}

/**
 * Checks if the process has a temporary access to the file.
 *
 * @param filename: The file that the process is trying to access
 * @pram pid: PID of the process
 * @return: 0 if access is denied, 1 if access is allowed
 */

int check_temp_access_noparent(const char *filename, pid_t pid) {
  // TODO: more efficient locking
  pthread_mutex_lock(&temp_permissions_table_lock);
  struct temp_process_permissions *permission_entry =
      get(&temp_permissions_table, pid);
  if (permission_entry != NULL) {
    unsigned long long process_creation_time = get_process_creation_time(pid);
    if (process_creation_time == 0) {
      perror("Could not retrieve process creation time");
      pthread_mutex_unlock(&temp_permissions_table_lock);
      return 0;
    }

    if (process_creation_time == permission_entry->creation_time) {
      // the process is the same as the one that was granted temporary access
      // to the file
      for_each(&permission_entry->allowed_files, allowed_file) {
        if (strncmp(*allowed_file, filename, strlen(filename)) == 0) {
          pthread_mutex_unlock(&temp_permissions_table_lock);
          return 1;
        }
      }
    }
  }
  pthread_mutex_unlock(&temp_permissions_table_lock);
  return 0;
}

/**
 * Finds the parent process ID of a given process.
 *
 * @param pid: The process ID of the process to find the parent of
 * @return: The parent process ID, or 0 if the parent process ID could not be
 * found
 */
pid_t get_parent_pid(pid_t pid) {
  pid_t ppid = 0;
  char path[256];
  snprintf(path, sizeof(path), "/proc/%u/status", pid);

  FILE *file = fopen(path, "r");
  if (file == NULL) {
    perror("Failed to open /proc/<pid>/status");
    return 0;
  }

  char line[256];
  while (fgets(line, sizeof(line), file)) {
    if (sscanf(line, "PPid:\t%d", &ppid) == 1) {
      fclose(file);
      return ppid;
    }
  }

  fclose(file);
  return 0; // Parent PID not found
}

/**
 * Checks if the process or any of it's parents have temporary access to the
 * file.
 *
 * @param filename: The file that the process is trying to access
 * @pram pi: The process information
 * @return: 0 if access is denied, 1 if access is allowed
 * @note: In case one of the parent processes is killed while this function
 * execution the result is not guranteed to be correct. It should only lead to
 * false negatives, though.
 */
int check_temp_access(const char *filename, struct process_info pi) {
  pid_t current_pid = pi.PID;
  while (current_pid != 0) {
    if (check_temp_access_noparent(filename, current_pid)) {
      return 1;
    }
    current_pid = get_parent_pid(current_pid);
  }

  return 0;
}

/**
 * Gives temporary access to the process to the file.
 *
 * @param filename: The file that the process is trying to access
 * @param pi: The process information
 * @return: 0 on success, -1 on failure (e.g. ENOMEM)
 */
int give_temp_access(const char *filename, struct process_info pi) {
  pthread_mutex_lock(&temp_permissions_table_lock);
  struct temp_process_permissions *permission_entry =
      get(&temp_permissions_table, pi.PID);

  if (permission_entry != NULL) {

    unsigned long long process_creation_time =
        get_process_creation_time(pi.PID);
    if (process_creation_time == 0) {
      perror("Could not retrieve process creation time");
      pthread_mutex_unlock(&temp_permissions_table_lock);
      return -1;
    }

    if (process_creation_time == permission_entry->creation_time) {
      // the process is the same as the one that was granted temporary access
      // to the file
      push(&permission_entry->allowed_files, strdup(filename));
      pthread_mutex_unlock(&temp_permissions_table_lock);
      return 0;
    }
    // we have an entry for the process, but the process is different
    // delete the entry and create a new one
    erase(&temp_permissions_table, pi.PID);
    permission_entry = NULL;
  }

  // no entry is present
  // construct the entry
  struct temp_process_permissions new_permission_entry;

  new_permission_entry.creation_time = get_process_creation_time(pi.PID);
  init(&new_permission_entry.allowed_files);
  push(&new_permission_entry.allowed_files, strdup(filename));

  insert(&temp_permissions_table, pi.PID, new_permission_entry);

  pthread_mutex_unlock(&temp_permissions_table_lock);
  return 0;
}
